#include "ehs/crt/CRT_LNX.h"
#include "ehs/Log.h"
#include "ehs/HRNG.h"

#include <sys/syscall.h>
#include <sys/mman.h>
#include <unistd.h>
#include <cstdarg>

extern "C"
{
	long syscall(long number, ...) {
		va_list args;
		va_start(args, number);

		// Retrieve the system call arguments.
		unsigned long arg1 = va_arg(args, unsigned long);
		unsigned long arg2 = va_arg(args, unsigned long);
		unsigned long arg3 = va_arg(args, unsigned long);
		unsigned long arg4 = va_arg(args, unsigned long);
		unsigned long arg5 = va_arg(args, unsigned long);
		unsigned long arg6 = va_arg(args, unsigned long);
		va_end(args);

		// Variable to store the system call return value
		long ret;

		// Inline assembly to perform the system call
		__asm__ volatile (
				"mov %1, %%rax\n\t" // System call number -> RAX
				"mov %2, %%rdi\n\t" // First argument -> RDI
				"mov %3, %%rsi\n\t" // Second argument -> RSI
				"mov %4, %%rdx\n\t" // Third argument -> RDX
				"mov %5, %%r10\n\t" // Fourth argument -> R10
				"mov %6, %%r8\n\t"  // Fifth argument -> R8
				"mov %7, %%r9\n\t"  // Sixth argument -> R9
				"syscall\n\t"       // Invoke system call
				"mov %%rax, %0\n\t" // Store return value
				: "=r"(ret) // output
				: "r"(number), "r"(arg1), "r"(arg2), "r"(arg3), "r"(arg4), "r"(arg5), "r"(arg6) // inputs
				: "%rax", "%rdi", "%rsi", "%rdx", "%r10", "%r8", "%r9", "memory", "cc" // clobbers
				);

		return ret;
	}

	int* __errno_location()
	{
		static __thread int tls_errno = 0; // __thread or thread_local (C++11 and later)
		return &tls_errno;
	}

	extern void (*__start_init_array[])();
	extern void (*__stop_init_array[])();
	extern void (*__start_fini_array[])();
	extern void (*__stop_fini_array[])();

	#ifdef EHS_DEBUG
		unsigned long __stack_chk_guard = 0x0000DEAD;
	#else
		unsigned long __stack_chk_guard = ehs::HRNG::GenerateSeed_u64();
	#endif

	void* __dso_handle = (void*)&__dso_handle;

	void __stack_chk_fail(void)
	{
		EHS_LOG("Error", 0, "Found corrupted stack.");

		syscall(SYS_exit, 1);
	}

	int __cxa_atexit(void (*func)(void *), void *arg, void * /*dso_handle*/)
	{
		return 0;
	}

	void construct()
	{
		for (void (**p)() = __start_init_array; p < __stop_init_array; ++p)
			(*p)();
	}

	void destruct()
	{
		for (void (**p)() = __start_fini_array; p < __stop_fini_array; ++p)
			(*p)();
	}

	size_t strlen(const char *s)
	{
		size_t count = 0;
		while (s[count])
			count++;

		return count;
	}
}

struct Block
{
	size_t size;
	Block* prev;
	Block* next;
	int free;
};

Block* first = nullptr;
Block* last = nullptr;

/// Extends heap memory upwards, towards zero.
/// @param [in] s The size of the memory needed aligned by 4 bytes.
/// @returns The new memory block.
Block* extend_heap(ehs::UInt_64 s)
{
	Block* b = (Block*)syscall(SYS_brk, nullptr);

	if ((void*)syscall(SYS_brk, (char*)(b + 1) + s) == (void*)-1)
		return nullptr;

	b->size = s;
	b->prev = last;
	b->next = nullptr;
	b->free = 0;

	if (last)
		last->next = b;

	last = b;

	return b;
}

/// Finds the first block that will fit the given size.
/// @param [in] s The 4 byte aligned size to look for.
/// @returns The matching available memory block.
Block* find_first(ehs::UInt_64 s)
{
	Block* current = first;
	while (current && (!current->free || current->size < s))
		current = current->next;

	return current;
}

/// Fragments an existing free memory block into the given size.
/// @param [in] in The memory block to fragment.
/// @param [in] s The size of the new memory block.
/// @returns The new memory block.
Block* fragment_block(Block* in, ehs::UInt_64 s)
{
	ehs::UInt_64 totalSize = sizeof(Block) + s;
	Block* b = (Block*)((char*)(in + 1) + (in->size - totalSize));
	b->size = s;
	b->prev = in;
	b->next = in->next;
	b->free = 0;

	in->size -= totalSize;
	in->next = b;

	return b;
}

/// Will find or allocate a memory block.
/// @param [in] size The size of the memory block to request.
/// @returns The requested memory on the heap.
void* malloc(ehs::UInt_64 size)
{
	Block* b;

	size = ((size - 1) >> 2 << 2) + 4;

	if (first)
	{
		b = find_first(size);
		if (!b)
			b = extend_heap(size);
		else if (b->size > sizeof(Block) + size)
			b = fragment_block(b, size);
	}
	else
	{
		b = extend_heap(size);
		if (!b)
			return nullptr;

		first = b;
	}

	return b + 1;
}

void* realloc(void* ptr, ehs::UInt_64 new_size)
{
	if (!ptr)
		return malloc(new_size);

	if (!new_size)
		return nullptr;

	Block* b = (Block*)ptr - 1;
	if (b->free)
		return nullptr;

	if (b->size == new_size)
		return ptr;
	else if (b->size > sizeof(Block) + new_size)
		return fragment_block(b, new_size) + 1;

	return nullptr;
}

/// Will flag the provided memory as free and will defragment other blocks adjacent to it.
/// @param [in] ptr The memory to flag as free.
/// @note If all data after the provided memory is free, it will reduce the heap size.
void free(void* ptr)
{
	if (!ptr)
		return;

	Block* b = (Block*)ptr - 1;
	if (b->free)
		return;

	b->free = 1;

	while (b->prev && b->prev->free)
		b = b->prev;

	while (b->next && b->next->free)
	{
		b->size += sizeof(Block) + b->next->size;
		b->next = b->next->next;
	}

	if (!b->next)
		syscall(SYS_brk, b);
}

ehs::UInt_64 get_heap_size()
{
	if (!first || !last)
		return 0;

	return ((char*)last + last->size) - (char*)first;
}

void* operator new(unsigned long size) noexcept
{
	return malloc(size);
}

void operator delete(void* ptr, unsigned long size) noexcept
{
	free(ptr);
}

void* operator new[](unsigned long size) noexcept(false)
{
	return malloc(size);
}

void operator delete[](void* ptr, unsigned long size) noexcept
{
	free(ptr);
}

void operator delete[](void* ptr) noexcept
{
	free(ptr);
}