#include "ehs/io/Console_LNX.h"
#include "ehs/io/File.h"
#include "ehs/Log.h"

#include <fcntl.h>
#include <unistd.h>
#include <cerrno>
#include <sys/syscall.h>

namespace ehs
{
	int Console::hdlOut = 0;
	int Console::hdlIn = 0;

	void Console::Attach()
	{
		hdlOut = (int)syscall(SYS_open, "/dev/stdout", O_WRONLY | O_SYNC);
		hdlIn = (int)syscall(SYS_open, "/dev/stdin", O_RDONLY);
	}

	void Console::Free()
	{
		int code = (int)syscall(SYS_close, hdlOut);
		if (code == -1)
			EHS_LOG_INT("Error", 0, "Failed to free the console output with error #" + Str_8::FromNum(errno) + ".");

		hdlOut = 0;

		code = (int)syscall(SYS_close, hdlIn);
		if (code == -1)
			EHS_LOG_INT("Error", 0, "Failed to free the console input with error #" + Str_8::FromNum(errno) + ".");

		hdlIn = 0;
	}

	void Console::Write_32(const Str_32& str, const bool newLine)
	{
		Str_32 result = str;
		if (newLine)
			result += U"\n";

		UInt_64 offset = 0;
		do {
			ssize_t written = syscall(SYS_write, hdlOut, &result[0], result.Size(true));
			if (written == -1)
			{
				EHS_LOG_INT("Error", 0, "Failed to write to console with error #" + Str_8::FromNum(errno) + ".");
				return;
			}
			offset += written;
		}
		while (offset < result.Size(true));
	}

	void Console::Write_16(const Str_16& str, const bool newLine)
	{
		Str_16 result = str;
		if (newLine)
			result += L"\n";

		UInt_64 offset = 0;
		do {
			ssize_t written = syscall(SYS_write, hdlOut, &result[0], result.Size(true));
			if (written == -1)
			{
				EHS_LOG_INT("Error", 0, "Failed to write to console with error #" + Str_8::FromNum(errno) + ".");
				return;
			}
			offset += written;
		}
		while (offset < result.Size(true));
	}

	void Console::Write_8(const Str_8& str, const bool newLine)
	{
		Str_8 result = str;
		if (newLine)
			result += "\n";

		UInt_64 offset = 0;
		do {
			ssize_t written = syscall(SYS_write, hdlOut, &result[0], result.Size());
			if (written == -1)
			{
				EHS_LOG_INT("Error", 0, "Failed to write to console with error #" + Str_8::FromNum(errno) + ".");
				return;
			}
			offset += written;
		}
		while (offset < result.Size());
	}

	Str_32 Console::Read_32(const UInt_64 bufferSize)
	{
		Str_32 result;
		Str_32 input(bufferSize);
		ssize_t read = 0;

		do
		{
			read = syscall(SYS_read, hdlIn, &input[0], bufferSize);
			if (read == -1)
			{
				EHS_LOG_INT("Error", 0, "Failed to read from console with error #" + Str_8::FromNum(errno) + ".");
				return result;
			}
			result.Push(input, read);
		}
		while (input[read - 1] != U'\n');

		return result;
	}

	Str_16 Console::Read_16(const UInt_64 bufferSize)
	{
		Str_16 result;
		Str_16 input(bufferSize);
		ssize_t read = 0;

		do
		{
			read = syscall(SYS_read, hdlIn, &input[0], bufferSize);
			if (read == -1)
			{
				EHS_LOG_INT("Error", 0, "Failed to read from console with error #" + Str_8::FromNum(errno) + ".");
				return result;
			}
			result.Push(input, read);
		}
		while (input[read - 1] != L'\n');

		return result;
	}

	Str_8 Console::Read_8(const UInt_64 bufferSize)
	{
		Str_8 result;
		Str_8 input(bufferSize);
		ssize_t read = 0;

		do
		{
			read = syscall(SYS_read, hdlIn, &input[0], bufferSize);
			if (read == -1)
			{
				EHS_LOG_INT("Error", 0, "Failed to read from console with error #" + Str_8::FromNum(errno) + ".");
				return result;
			}
			result.Push(input, read);
		}
		while (input[read - 1] != '\n');

		return result.Sub(0, result.Size() - 1);
	}

	void Console::Clear()
	{
		const Char_8 code[] = "\033[2J\033[1;1H";
		UInt_64 offset = 0;
		do {
			ssize_t written = syscall(SYS_write, hdlOut, code, sizeof(code));
			if (written == -1)
			{
				EHS_LOG_INT("Error", 0, "Failed to clear console with error #" + Str_8::FromNum(errno) + ".");
				return;
			}
			offset += written;
		}
		while (offset < sizeof(code));
	}

	void Console::SetTitle_32(const Str_32& title)
	{
		Str_32 code = U"\033]0;" + title + U"\007";
		UInt_64 offset = 0;
		do {
			ssize_t written = syscall(SYS_write, hdlOut, &code[0], code.Size(true));
			if (written == -1)
			{
				EHS_LOG_INT("Error", 0, "Failed to set console title with error #" + Str_8::FromNum(errno) + ".");
				return;
			}
			offset += written;
		}
		while (offset < code.Size(true));
	}

	void Console::SetTitle_16(const Str_16& title)
	{
		Str_16 code = L"\033]0;" + title + L"\007";
		UInt_64 offset = 0;
		do {
			ssize_t written = syscall(SYS_write, hdlOut, &code[0], code.Size(true));
			if (written == -1)
			{
				EHS_LOG_INT("Error", 0, "Failed to set console title with error #" + Str_8::FromNum(errno) + ".");
				return;
			}
			offset += written;
		}
		while (offset < code.Size(true));
	}

	void Console::SetTitle_8(const Str_8& title)
	{
		Str_8 code = "\033]0;" + title + "\007";
		UInt_64 offset = 0;
		do {
			ssize_t written = syscall(SYS_write, hdlOut, &code[0], code.Size());
			if (written == -1)
			{
				EHS_LOG_INT("Error", 0, "Failed to set console title with error #" + Str_8::FromNum(errno) + ".");
				return;
			}
			offset += written;
		}
		while (offset < code.Size());
	}

	Str_32 Console::GetTitle_32()
	{
		return {};
	}

	Str_16 Console::GetTitle_16()
	{
		return {};
	}

	Str_8 Console::GetTitle_8()
	{
		return {};
	}

	Vector<Str_32> Console::GetArgs_32(const UInt_64 bufferSize)
	{
		File cmdFile("/proc/self/cmdline", Mode::READ, Disposition::OPEN);
		Array<Byte> data = cmdFile.ReadArray(bufferSize);
		cmdFile.Release();

		Vector<Str_32> args;

		UInt_64 o = 0;
		for (UInt_64 i = 0; i < data.Size() - 1; i += sizeof(Char_8))
		{
			if (data[i] != '\0')
				continue;



			args.Push(UTF::To_32((Char_8*)&data[o], i - o));
			o = i + sizeof(Char_8);
		}

		if (o < data.Size())
			args.Push(UTF::To_32((Char_8*)&data[o], data.Size() - o - 1));

		return args;
	}

	Vector<Str_16> Console::GetArgs_16(const UInt_64 bufferSize)
	{
		File cmdFile("/proc/self/cmdline", Mode::READ, Disposition::OPEN);
		Array<Byte> data = cmdFile.ReadArray(bufferSize);
		cmdFile.Release();

		Vector<Str_16> args;

		UInt_64 o = 0;
		for (UInt_64 i = 0; i < data.Size() - 1; i += sizeof(Char_8))
		{
			if (data[i] != '\0')
				continue;

			args.Push(UTF::To_16((Char_8*)&data[o], i - o));
			o = i + sizeof(Char_8);
		}

		if (o < data.Size())
			args.Push(UTF::To_16((Char_8*)&data[o], data.Size() - o - 1));

		return args;
	}

	Vector<Str_8> Console::GetArgs_8(const UInt_64 bufferSize)
	{
		File cmdFile("/proc/self/cmdline", Mode::READ, Disposition::OPEN);
		Array<Byte> data = cmdFile.ReadArray(bufferSize);
		cmdFile.Release();

		Vector<Str_8> args;

		UInt_64 o = 0;
		for (UInt_64 i = 0; i < data.Size() - 1; i += sizeof(Char_8))
		{
			if (data[i] != '\0')
				continue;

			args.Push(Str_8((Char_8*)&data[o], i - o));
			o = i + sizeof(Char_8);
		}

		if (o < data.Size())
			args.Push(Str_8((Char_8*)&data[o], data.Size() - o - 1));

		return args;
	}
}