#include "ehs/system/Thread_LNX.h"
#include "ehs/Serializer.h"

#include <sys/mman.h>
#include <sys/syscall.h>
#include <sys/wait.h>
#include <sys/errno.h>
#include <pthread.h>

namespace ehs
{
	Thread::~Thread()
	{
		if (detached && !*detached)
		{
			if (stack && syscall(SYS_munmap, stack, GetStackSize()) == -1)
				EHS_LOG_INT("Error", 0, "Failed to deallocate stack for thread with error #" + Str_8::FromNum(errno) + ".");

			delete detached;
		}
	}

	Thread::Thread()
		: BaseThread(0), stack(nullptr), hdl(0), detached(nullptr)
	{
	}

	Thread::Thread(UInt_64 stackSize, void* args, UInt_32 (*func)(void*))
		: BaseThread(1024 * stackSize), stack(nullptr), hdl(0), detached(new bool(false))
	{
		stack = (void*)syscall(SYS_mmap, nullptr, GetStackSize(), PROT_READ | PROT_WRITE, MAP_PRIVATE | MAP_ANONYMOUS, -1, 0);
		if (stack == MAP_FAILED)
		{
			EHS_LOG_INT("Error", 0, "Failed to allocate stack for thread.");
			return;
		}

		Serializer<UInt_64>* data = new Serializer<UInt_64>(Endianness::LE);
		data->Write(GetStackSize());
		data->Write(stack);
		data->Write(detached);
		data->Write(args);
		data->Write(func);

		hdl = (pid_t)syscall(SYS_clone, Redirect, stackSize, CLONE_VM | CLONE_FS | CLONE_FILES | CLONE_SIGHAND | CLONE_THREAD, data);
		id = hdl;
	}

	Thread::Thread(Thread&& thread) noexcept
		: BaseThread((BaseThread&&)thread), stack(thread.stack), hdl(thread.hdl), detached(thread.detached)
	{
		thread.stack = nullptr;
		thread.hdl = 0;
		thread.detached = nullptr;
	}

	Thread::Thread(const Thread& thread)
		: BaseThread(thread), stack(nullptr), hdl(0), detached(nullptr)
	{
	}

	Thread& Thread::operator=(Thread&& thread) noexcept
	{
		if (this == &thread)
			return *this;

		Join();

		BaseThread::operator=((BaseThread&&)thread);

		stack = thread.stack;
		hdl = thread.hdl;
		detached = thread.detached;

		thread.stack = nullptr;
		thread.hdl = 0;
		thread.detached = nullptr;

		return *this;
	}

	Thread& Thread::operator=(const Thread& thread)
	{
		if (this == &thread)
			return *this;

		Join();

		BaseThread::operator=(thread);

		stack = nullptr;
		hdl = 0;
		detached = nullptr;

		return *this;
	}

	void Thread::Detach()
	{
		BaseThread::Detach();

		stack = nullptr;
		*detached = true;
		detached = nullptr;
		hdl = 0;
	}

	void Thread::Join()
	{
		if (detached && *detached)
			return;

		syscall(SYS_wait4, hdl, nullptr, 0, nullptr);
		hdl = 0;

		if (syscall(SYS_munmap, stack, GetStackSize()) == -1)
			EHS_LOG_INT("Error", 0, "Failed to deallocate stack in detached thread with error #" + Str_8::FromNum(errno) + ".");

		stack = nullptr;

		delete detached;
		detached = nullptr;
	}

	UInt_32 Thread::GetCurrentId()
	{
		return syscall(SYS_gettid);
	}

	void Thread::HardSleep(UInt_32 milliseconds)
	{
		timespec req = {milliseconds / 1000, milliseconds % 1000 * 1000000};
		syscall(SYS_nanosleep, &req, nullptr);
	}

	UInt_32 Thread::Redirect(void* args)
	{
		Serializer<UInt_64>* data = (Serializer<UInt_64>*)args;
		UInt_64 stackSize = data->Read<UInt_64>();
		void* stack = data->Read<void*>();
		bool* detached = data->Read<bool*>();
		void* inArgs = data->Read<void*>();
		UInt_32 (*func)(void*) = data->Read<UInt_32 (*)(void*)>();
		delete data;

		UInt_32 error = func(inArgs);

		if (*detached)
		{
			if (syscall(SYS_munmap, stack, stackSize) == -1)
			{
				error = 1;
				EHS_LOG_INT("Error", 0, "Failed to deallocate stack in detached thread with error #" + Str_8::FromNum(errno) + ".");
			}

			delete detached;
		}

		return error;
	}
}